@@ -1,5 +1,6 @@
+#!/usr/bin/env python3
 import os
-from typing import Tuple
+from typing import Tuple, Union
 import numpy as np
 from astropy.coordinates import SkyCoord
 import astropy.units as u
@@ -13,16 +14,18 @@ import corner
 from astropy.visualization import quantity_support
 from astropy.modeling import models, fitting
 import pandas as pd
+import numba as nb
+import xarray as xr
+import logging as logger
+
+logger.basicConfig(
+    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
+    datefmt="%Y-%m-%d %H:%M:%S",
+    force=True,
+)
 
 quantity_support()
 warnings.filterwarnings("ignore")
-# warnings.filterwarnings("ignore", message="All-NaN slice encountered")
-# warnings.filterwarnings("ignore", message="All-NaN axis encountered")
-# warnings.filterwarnings("ignore", message="Mean of empty slice")
-# warnings.filterwarnings("ignore", message="converting a masked element to nan")
-# warnings.filterwarnings("ignore", message="invalid value encountered in power")
-# warnings.filterwarnings("ignore", message="overflow encountered in power")
-
 
 def model(x, amplitude, x_break, alpha_1, alpha_2):
     alpha = np.where(x < x_break, alpha_1, alpha_2)
@@ -36,7 +39,7 @@ def astropy_fit(x, y, y_err, y_dist, outdir, label, verbose=False, **kwargs):
 
     # initialize a linear model
     line_init = models.BrokenPowerLaw1D()
-    fitted_line = fitter(line_init, x, y, weights=1.0 / y_err ** 2)
+    fitted_line = fitter(line_init, x, y, weights=1.0 / y_err**2)
     amplitude, x_break, alpha_1, alpha_2 = fitted_line.parameters
     posterior = {
         "amplitude": amplitude,
@@ -100,13 +103,46 @@ def bilby_fit(x, y, y_err, y_dist, outdir, label, verbose=False, **kwargs):
     return result
 
 
+def combinate(data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
+    """Return all combinations of data with itself
+
+    Args:
+        data (np.ndarray): Data to combine.
+
+    Returns:
+        Tuple[np.ndarray, np.ndarray]: Data_1 matched with Data_2
+    """
+    ix, iy = np.triu_indices(data.shape[0], k=1)
+    idx = np.vstack((ix, iy)).T
+    dx, dy = data[idx].swapaxes(0, 1)
+    return dx, dy
+
+
+@nb.njit(parallel=True)
+def mc_sample(data: np.ndarray, errors: np.ndarray, samples: int = 1000) -> np.ndarray:
+    """Sample errors using Monte-Carlo
+    Assuming Gaussian distribution.
+
+    Args:
+        data (np.ndarray): Measurements
+        errors (np.ndarray): Errors
+        samples (int, optional): Samples of the distribution. Defaults to 1000.
+
+    Returns:
+        np.ndarray: Sample array. Shape (len(data/errors),samples)
+    """
+    data_dist = np.zeros((len(data), samples)).astype(data.dtype)
+    for i in nb.prange(data.shape[0]):
+        data_dist[i] = np.random.normal(loc=data[i], scale=errors[i], size=samples)
+    return data_dist
+
+
 def structure_function(
     data: u.Quantity,
     errors: u.Quantity,
     coords: SkyCoord,
     samples: int,
-    bins: u.Quantity,
-    weights: np.ndarray = None,
+    bins: Union[u.Quantity, int],
     show_plots: bool = False,
     save_plots: bool = False,
     verbose: bool = False,
@@ -122,7 +158,7 @@ def structure_function(
         errors (u.Quantity): 1D array of errors.
         coords (SkyCoord): 1D array of coordinates.
         samples (int): Number of samples to use for Monte-Carlo error propagation.
-        bins (u.Quantity): Bin edges of the structure function.
+        bins (Union[u.Quantity, int]): Bin edges of the structure function, or number of bins.
         show_plots (bool, optional): Show plots. Defaults to False.
         verbose (bool, optional): Print progress. Defaults to False.
         fit (str, optional): How to fit the broken powerlaw. Can be 'astropy', 'astropy_mc' or 'bilby'. Defaults to None.
@@ -130,98 +166,114 @@ def structure_function(
         **kwargs: Additional keyword arguments to pass to the bilby.core.run_sampler function.
 
     Returns:
-        Tuple[u.Quantity, u.Quantity, Tuple[u.Quantity, u.Quantity], np.ndarray]: 
+        Tuple[u.Quantity, u.Quantity, Tuple[u.Quantity, u.Quantity], np.ndarray]:
             cbins: center of bins.
             medians: median of the structure function.
             errors: upper and lower error of the structure function.
             counts: number of source pairs in each bin.
     """
 
-    # Sample the errors assuming a Gaussian distribution
     if verbose:
-        print("Sampling errors...")
-    rm_dist = []
-    d_rm_dist = []
-    for i in tqdm(range(data.shape[0]), "Sample Gaussian", disable=not verbose):
-        rm_dist.append(
-            np.random.normal(loc=data[i].value, scale=errors[i].value, size=samples)
+        logger.basicConfig(
+            level=logger.INFO,
+            format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
+            datefmt="%Y-%m-%d %H:%M:%S",
+            force=True,
         )
-        d_rm_dist.append(
-            np.random.normal(loc=errors[i].value, scale=errors[i].value, size=samples)
-        )
-    rm_dist = np.array(rm_dist)
-    d_rm_dist = np.array(d_rm_dist)
+
+    # Sample the errors assuming a Gaussian distribution
+
+    logger.info("Sampling errors...")
+    rm_dist = mc_sample(
+        data=data.value,
+        errors=errors.value,
+        samples=samples,
+    )
+    d_rm_dist = mc_sample(
+        data=errors.value,
+        errors=errors.value,
+        samples=samples,
+    )
 
     # Get all combinations of sources and compute the difference
-    if verbose:
-        print("Getting data differences...")
-    F_dist = np.array(list(itertools.combinations(rm_dist, r=2)))
-    if weights is None:
-        weights = np.ones(data.shape[0])
-    w_dist = np.mean(np.array(list(itertools.combinations(weights, r=2))), axis=1)
-    diffs_dist = np.transpose(np.power(np.subtract(F_dist[:, 0], F_dist[:, 1]), 2,))
+
+    logger.info("Getting data differences...")
+    diffs_dist = np.subtract(*combinate(rm_dist)).T ** 2
 
     # Get all combinations of data_errs sources and compute the difference
-    if verbose:
-        print("Getting data error differences...")
-    dF_dist = np.array(list(itertools.combinations(d_rm_dist, r=2)))
-    d_diffs_dist = np.transpose(np.power(np.subtract(dF_dist[:, 0], dF_dist[:, 1]), 2,))
+
+    logger.info("Getting data error differences...")
+    d_diffs_dist = np.subtract(*combinate(d_rm_dist)).T ** 2
 
     # Get the angular separation of the source pairs
-    if verbose:
-        print("Getting angular separations...")
-    cx_ra_perm, cy_ra_perm = np.array(
-        list(itertools.combinations(coords.ra.to(u.deg).value, r=2))
-    ).T
-    cx_dec_perm, cy_dec_perm = np.array(
-        list(itertools.combinations(coords.dec.to(u.deg).value, r=2))
-    ).T
-    coords_x = SkyCoord(cx_ra_perm * u.deg, cx_dec_perm * u.deg)
-    coords_y = SkyCoord(cy_ra_perm * u.deg, cy_dec_perm * u.deg)
-    dtheta = coords_x.separation(coords_y)
 
+    logger.info("Getting angular separations...")
+    ra_1, ra_2 = combinate(coords.ra)
+    dec_1, dec2 = combinate(coords.dec)
+
+    coords_1 = SkyCoord(ra_1, dec_1)
+    coords_2 = SkyCoord(ra_2, dec2)
+    dtheta = coords_1.separation(coords_2)
+
+    # Auto compute bins
+    if type(bins) is int:
+
+        logger.info("Auto-computing bins...")
+        nbins = bins
+        start = np.log10(np.min(dtheta).to(u.deg).value)
+        stop = np.log10(np.max(dtheta).to(u.deg).value)
+        bins = np.logspace(start, stop, nbins, endpoint=True) * u.deg
+    else:
+        nbins = len(bins)
     # Compute the SF
-    if verbose:
-        print("Computing SF...")
-    sf_dists = np.zeros((len(bins) - 1, samples)) * np.nan
-    d_sf_dists = np.zeros((len(bins) - 1, samples)) * np.nan
-    count = np.zeros((len(bins) - 1)) * np.nan
-    cbins = np.zeros((len(bins) - 1)) * np.nan * u.deg
-    for i, b in enumerate(tqdm(bins[:-1], disable=not verbose)):
-        bin_idx = (bins[i] <= dtheta) & (dtheta < bins[i + 1])
-        centre = (bins[i] + bins[i + 1]) / 2
 
-        cbins[i] = centre
-        count[i] = np.sum(bin_idx)
-        try:
-            sf_dist = np.average(
-                diffs_dist[:, bin_idx], axis=1, weights=w_dist[bin_idx]
-            )
-            d_sf_dist = np.average(
-                d_diffs_dist[:, bin_idx], axis=1, weights=w_dist[bin_idx]
-            )
-            sf_dists[i] = sf_dist
-            d_sf_dists[i] = d_sf_dist
-        except ZeroDivisionError:
-            continue
+    logger.info("Computing SF...")
+    bins_idx = pd.cut(dtheta, bins, include_lowest=True, labels=False, right=True).astype(int)
+    cbins = np.sqrt(bins[1:] * bins[:-1])  # Take geometric mean of bins - assuming log
 
+    diffs_xr = xr.Dataset(
+        dict(
+            data=(["samples", "source pair"], diffs_dist),
+            error=(["samples", "source pair"], d_diffs_dist),
+        ),
+        coords=dict(
+            bins_idx=("source pair", bins_idx),
+        ),
+    )
+
+    # Compute SF
+    sf_xr = diffs_xr.groupby("bins_idx").mean(dim="source pair")
+    count_xr = diffs_xr["bins_idx"].groupby("bins_idx").count()
     # Get the final SF correcting for the errors
-    sf_dists_cor = sf_dists - d_sf_dists
-    medians = np.nanmedian(sf_dists_cor, axis=1)
-    per16 = np.nanpercentile(sf_dists_cor, 16, axis=1)
-    per84 = np.nanpercentile(sf_dists_cor, 84, axis=1)
+    sf_xr_cor = sf_xr.data - sf_xr.error
+    per16_xr, medians_xr, per84_xr = sf_xr_cor.quantile(
+        (0.16, 0.5, 0.84), dim="samples"
+    )
+    # Return to numpy arrays for use later
+    count = np.zeros_like(cbins.value)
+    medians = np.zeros_like(cbins.value)
+    per16 = np.zeros_like(cbins.value)
+    per84 = np.zeros_like(cbins.value)
+    sf_dists_cor = np.zeros((len(cbins), samples))
+    sf_dists = np.zeros((len(cbins), samples))
+    d_sf_dists = np.zeros((len(cbins), samples))
+    for arr, xarr in zip(
+        (count, medians, per16, per84, sf_dists_cor, sf_dists, d_sf_dists),
+        (count_xr, medians_xr, per16_xr, per84_xr, sf_xr_cor, sf_xr.data, sf_xr.error),
+    ):
+        idx = count_xr.coords.to_index()
+        nan_clip = idx>=0 # Clip NaN indices
+        arr[idx[nan_clip]] = xarr[nan_clip]
     err_low = medians - per16
     err_high = per84 - medians
     err = np.array([err_low.astype(float), err_high.astype(float)])
-
+    if outdir is None:
+        outdir = "outdir"
+    bilby.utils.check_directory_exists_and_if_not_mkdir(outdir)
     if fit:
-        if verbose:
-            print("Fitting SF with a broken power law...")
+        logger.info("Fitting SF with a broken power law...")
         # A few simple setup steps
         label = "linear_regression"
-        if outdir is None:
-            outdir = "outdir"
-        bilby.utils.check_directory_exists_and_if_not_mkdir(outdir)
 
         # Only use bins with at least 10 sources
         cut = (
@@ -246,13 +298,17 @@ def structure_function(
             raise ValueError("Invalid fit type")
         result = fit_func(x, y, y_err, y_dist, outdir, label, verbose=verbose, **kwargs)
         if show_plots and fit != "astropy":
+            try:
+                result.plot_corner(dpi=300, save=save_plots)
+            except:
+                pass
             samps = result.samples
             labels = result.parameter_labels
-            fig = plt.figure(figsize=(10, 10), facecolor="w")
+            fig = plt.figure(facecolor="w")
             fig = corner.corner(samps, labels=labels, fig=fig)
             if save_plots:
-                plt.savefig(os.path.join(outdir, "corner.pdf"))
-        if verbose:
+                plt.savefig(os.path.join(outdir, "corner.pdf"), dpi=300, bbox_inches="tight")
+
             amp_ps = np.nanpercentile(result.posterior["amplitude"], [16, 50, 84])
             break_ps = np.nanpercentile(result.posterior["x_break"], [16, 50, 84])
             a1_ps = np.nanpercentile(result.posterior["alpha_1"], [16, 50, 84])
@@ -263,17 +319,17 @@ def structure_function(
             alpha_1 = a1_ps[1]
             alpha_2 = a2_ps[1]
 
-            if fit != "astropy": 
+            if fit != "astropy":
                 amplitude = round(amplitude, uncertainty=amp_ps[2] - amp_ps[1])
                 x_break = round(x_break, uncertainty=break_ps[2] - break_ps[1])
                 alpha_1 = round(alpha_1, uncertainty=a1_ps[2] - a1_ps[1])
                 alpha_2 = round(alpha_2, uncertainty=a2_ps[2] - a2_ps[1])
 
-            print("Fitting results:")
-            print(f"    Amplitude: {amplitude} [{data.unit**2}]")
-            print(f"    Break point: {x_break} [{u.deg}]")
-            print(f"    alpha 1 (theta < break): {alpha_1}")
-            print(f"    alpha 2 (theta > break): {alpha_2}")
+            logger.info("Fitting results:")
+            logger.info(f"    Amplitude: {amplitude} [{data.unit**2}]")
+            logger.info(f"    Break point: {x_break} [{u.deg}]")
+            logger.info(f"    alpha 1 (theta < break): {alpha_1}")
+            logger.info(f"    alpha 2 (theta > break): {alpha_2}")
     else:
         result = None
 
@@ -282,7 +338,7 @@ def structure_function(
     ##############################################################################
     if show_plots:
         good_idx = count >= 10
-        plt.figure(figsize=(6, 6), facecolor="w")
+        plt.figure(facecolor="w")
         plt.plot(
             cbins[good_idx],
             medians[good_idx],
@@ -354,9 +410,9 @@ def structure_function(
         plt.ylim(np.nanmin(medians) / 10, np.nanmax(medians) * 10)
         plt.legend()
         if save_plots:
-            plt.savefig(os.path.join(outdir, "errorbar.png"))
+            plt.savefig(os.path.join(outdir, "errorbar.pdf"), dpi=300, bbox_inches="tight")
 
-        plt.figure(figsize=(6, 6), facecolor="w")
+        plt.figure(facecolor="w")
         plt.plot(cbins, count, ".", color="tab:red", label="Median from MC")
         plt.xscale("log")
         plt.yscale("log")
@@ -364,12 +420,12 @@ def structure_function(
         plt.ylabel(r"Number of source pairs")
         plt.xlim(bins[0].value, bins[-1].value)
         if save_plots:
-            plt.savefig(os.path.join(outdir, "counts.png"))
+            plt.savefig(os.path.join(outdir, "counts.pdf"), dpi=300, bbox_inches="tight")
 
         counts = []
         cor_dists = sf_dists - d_sf_dists
-        plt.figure()
-        for dist in tqdm(cor_dists, disable=not verbose):
+        plt.figure(facecolor="w")
+        for dist in tqdm(cor_dists, disable=not verbose, desc="Making hist plot"):
             n, hbins, _ = plt.hist(
                 dist, range=(np.nanmin(cor_dists), np.nanmax(cor_dists)), bins=100
             )
@@ -385,8 +441,8 @@ def structure_function(
         x = cbins
         y = c_hbins
         X, Y = np.meshgrid(x, y)
-        plt.figure(figsize=(7, 6), facecolor="w")
-        plt.pcolormesh(X, Y, counts.T, cmap=plt.cm.cubehelix_r, shading="auto")
+        plt.figure(facecolor="w")
+        plt.pcolormesh(X, Y, counts.T, cmap=plt.cm.cubehelix_r)
         plt.colorbar()
         plt.xticks(x)
         plt.yticks(y)
@@ -409,7 +465,7 @@ def structure_function(
             label="Expected saturation ($2\sigma^2$)",
         )
         if save_plots:
-            plt.savefig(os.path.join(outdir, "PDF.png"))
+            plt.savefig(os.path.join(outdir, "PDF.pdf"), dpi=300, bbox_inches="tight")
     ##############################################################################
 
     return cbins, medians, err, count, result
